searchNodes=[{"doc":"Bird View Finitomata provides a boilerplate for FSM implementation, allowing to concentrate on the business logic rather than on the process management and transitions/events consistency tweaking. It reads a description of the FSM from a string in PlantUML , Mermaid , or even custom format. Basically, it looks more or less like this PlantUML [ * ] -- &gt; s1 : to_s1 s1 -- &gt; s2 : to_s2 s1 -- &gt; s3 : to_s3 s2 -- &gt; [ * ] : ok s3 -- &gt; [ * ] : ok Mermaid s1 -- &gt; | to_s2 | s2 s1 -- &gt; | to_s3 | s3 Note mermaid does not allow to explicitly specify transitions (and hence event names) from the starting state and to the end state(s), these states names are implicitly set to :* and events to :__start__ and :__end__ respectively. Finitomata validates the FSM is consistent, namely it has a single initial state, one or more final states, and no orphan states. If everything is OK, it generates a GenServer that could be used both alone, and with provided supervision tree. This GenServer requires to implement three callbacks on_transition/4 — mandatory on_failure/3 — optional on_enter/2 — optional on_exit/2 — optional on_terminate/1 — optional on_timer/2 — optional All the callbacks do have a default implementation, that would perfectly handle transitions having a single to state and not requiring any additional business logic attached. Upon start, it moves to the next to initial state and sits there awaiting for the transition request . Then it would call an on_transition/4 callback and move to the next state, or remain in the current one, according to the response. Upon reaching a final state, it would terminate itself. The process keeps all the history of states it went through, and might have a payload in its state. Special Events If the event name is ended with a bang (e. g. idle --&gt; |start!| started ) and this event is the only one allowed from this state (there might be several transitions though,) it’d be considered as determined and FSM will be transitioned into the new state instantly. If the event name is ended with a question mark (e. g. idle --&gt; |start?| started ,) the transition is considered as expected to fail; no on_failure/2 callback would be called on failure and no log warning will be printed. FSM Tuning and Configuration Recurrent Callback If timer: non_neg_integer() option is passed to use Finitomata , then Finitomata.on_timer/2 callback will be executed recurrently. This might be helpful if FSM needs to update its state from the outside world on regular basis. Automatic FSM Termination If auto_terminate: true() | state() | [state()] option is passed to use Finitomata , the special __end__ event to transition to the end state will be called automatically under the hood, if the current state is either listed explicitly, or if the value of the parameter is true . Ensuring State Entry If ensure_entry: true() | [state()] option is passed to use Finitomata , the transition attempt will be retried with {:continue, {:transition, {event(), event_payload()}}} message until succeeded. Neither on_failure/2 callback is called nor warning message is logged. The payload would be updated to hold __retries__: pos_integer() key. If the payload was not a map, it will be converted to a map %{payload: payload} . Example Let’s define the FSM instance defmodule MyFSM do @plantuml &quot;&quot;&quot; [*] --&gt; s1 : to_s1 s1 --&gt; s2 : to_s2 s1 --&gt; s3 : to_s3 s2 --&gt; [*] : ok s3 --&gt; [*] : ok &quot;&quot;&quot; use Finitomata , fsm : @plantuml , syntax : Finitomata.PlantUML ## or uncomment lines below for Mermaid syntax # @mermaid &quot;&quot;&quot; # s1 --&gt; |to_s2| s2 # s1 --&gt; |to_s3| s3 # &quot;&quot;&quot; # use Finitomata, fsm: @mermaid, syntax: Finitomata.Mermaid @impl Finitomata def on_transition ( :s1 , :to_s2 , event_payload , state_payload ) , do : { :ok , :s2 , state_payload } end Now we can play with it a bit. children = [ Finitomata . child_spec ( ) ] Supervisor . start_link ( children , strategy : :one_for_one ) Finitomata . start_fsm MyFSM , &quot;My first FSM&quot; , %{ foo : :bar } Finitomata . transition &quot;My first FSM&quot; , { :to_s2 , nil } Finitomata . state &quot;My first FSM&quot; #⇒ %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}} Finitomata . allowed? &quot;My first FSM&quot; , :* # state #⇒ true Finitomata . responds? &quot;My first FSM&quot; , :to_s2 # event #⇒ false Finitomata . transition &quot;My first FSM&quot; , { :ok , nil } # to final state #⇒ [info] [◉ ⇄] [state: %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}}] Finitomata . alive? &quot;My first FSM&quot; #⇒ false Typically, one would implement all the on_transition/4 handlers, pattern matching on the state/event.","ref":"Finitomata.html","title":"Finitomata","type":"behaviour"},{"doc":"Returns true if the supervision tree is alive, false otherwise.","ref":"Finitomata.html#alive?/0","title":"Finitomata.alive?/0","type":"function"},{"doc":"Returns true if the FSM specified is alive, false otherwise.","ref":"Finitomata.html#alive?/1","title":"Finitomata.alive?/1","type":"function"},{"doc":"Returns true if the transition to the state state is possible, false otherwise.","ref":"Finitomata.html#allowed?/2","title":"Finitomata.allowed?/2","type":"function"},{"doc":"This callback will be called on entering the state.","ref":"Finitomata.html#c:on_enter/2","title":"Finitomata.on_enter/2","type":"callback"},{"doc":"This callback will be called on exiting the state.","ref":"Finitomata.html#c:on_exit/2","title":"Finitomata.on_exit/2","type":"callback"},{"doc":"This callback will be called if the transition failed to complete to allow the consumer to take an action upon failure.","ref":"Finitomata.html#c:on_failure/3","title":"Finitomata.on_failure/3","type":"callback"},{"doc":"This callback will be called on transition to the final state to allow the consumer to perform some cleanup, or like.","ref":"Finitomata.html#c:on_terminate/1","title":"Finitomata.on_terminate/1","type":"callback"},{"doc":"This callback will be called recurrently if timer: pos_integer() option has been given to use Finitomata .","ref":"Finitomata.html#c:on_timer/2","title":"Finitomata.on_timer/2","type":"callback"},{"doc":"This callback will be called from each transition processor.","ref":"Finitomata.html#c:on_transition/4","title":"Finitomata.on_transition/4","type":"callback"},{"doc":"Returns true if the transition by the event event is possible, false otherwise.","ref":"Finitomata.html#responds?/2","title":"Finitomata.responds?/2","type":"function"},{"doc":"Starts the FSM instance. The arguments are the implementation of FSM (the module, having use Finitomata ) the name of the FSM (might be any term, but it must be unique) the payload to be carried in the FSM state during the lifecycle The FSM is started supervised.","ref":"Finitomata.html#start_fsm/3","title":"Finitomata.start_fsm/3","type":"function"},{"doc":"The state of the FSM.","ref":"Finitomata.html#state/1","title":"Finitomata.state/1","type":"function"},{"doc":"Initiates the transition. The arguments are the name of the FSM {event, event_payload} tuple; the payload will be passed to the respective on_transition/4 call delay (optional) the interval in milliseconds to apply transition after","ref":"Finitomata.html#transition/3","title":"Finitomata.transition/3","type":"function"},{"doc":"The payload that can be passed to each call to transition/3","ref":"Finitomata.html#t:event_payload/0","title":"Finitomata.event_payload/0","type":"type"},{"doc":"The name of the FSM (might be any term, but it must be unique)","ref":"Finitomata.html#t:fsm_name/0","title":"Finitomata.fsm_name/0","type":"type"},{"doc":"Error types of FSM validation","ref":"Finitomata.html#t:validation_error/0","title":"Finitomata.validation_error/0","type":"type"},{"doc":"Carries the state of the FSM.","ref":"Finitomata.State.html","title":"Finitomata.State","type":"module"},{"doc":"The payload that has been passed to the FSM instance on startup","ref":"Finitomata.State.html#t:payload/0","title":"Finitomata.State.payload/0","type":"type"},{"doc":"The internal representation of the FSM state","ref":"Finitomata.State.html#t:t/0","title":"Finitomata.State.t/0","type":"type"},{"doc":"The internal representation of Transition . It includes from and to states, and the event , all represented as atoms.","ref":"Finitomata.Transition.html","title":"Finitomata.Transition","type":"module"},{"doc":"Returns the list of all the transitions, matching the options. Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.Mermaid . parse ( ...&gt; &quot;idle --&gt; |to_s1| s1 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s2| s2 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s3| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |to_s3| s3&quot; ) ...&gt; Finitomata.Transition . allowed ( transitions , to : [ :idle , :* ] ) [ { :* , :idle , :__start__ } , { :s3 , :* , :__end__ } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s1 ) [ { :s1 , :s2 , :to_s2 } , { :s1 , :s3 , :to_s3 } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s1 , to : :s3 ) [ { :s1 , :s3 , :to_s3 } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s1 , with : :to_s3 ) [ { :s1 , :s3 , :to_s3 } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s2 , with : :to_s2 ) [ ]","ref":"Finitomata.Transition.html#allowed/2","title":"Finitomata.Transition.allowed/2","type":"function"},{"doc":"Returns the list of all the transitions, matching the from state and the event . Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . allowed ( transitions , :s1 , :foo ) [ :s2 ] ...&gt; Finitomata.Transition . allowed ( transitions , :s1 , :* ) [ ]","ref":"Finitomata.Transition.html#allowed/3","title":"Finitomata.Transition.allowed/3","type":"function"},{"doc":"Returns true if the transition from → to is allowed, false otherwise. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . allowed? ( transitions , :s1 , :s2 ) true ...&gt; Finitomata.Transition . allowed? ( transitions , :s1 , :* ) false","ref":"Finitomata.Transition.html#allowed?/3","title":"Finitomata.Transition.allowed?/3","type":"function"},{"doc":"Returns keyword list of {Finitomata.Transition.state(), [Finitomata.Transition.event()]} for transitions which do not have a determined to state. Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.Mermaid . parse ( ...&gt; &quot;idle --&gt; |to_s1| s1 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s2| s2 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s3| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |to_s1| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |ambiguous| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |ambiguous| s4 \\n &quot; &lt;&gt; ...&gt; &quot;s3 --&gt; |determined| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s3 --&gt; |determined| s4 \\n &quot; ) ...&gt; Finitomata.Transition . ambiguous ( transitions ) [ s3 : { :determined , [ :s3 , :s4 ] } , s2 : { :ambiguous , [ :s3 , :s4 ] } ]","ref":"Finitomata.Transition.html#ambiguous/1","title":"Finitomata.Transition.ambiguous/1","type":"function"},{"doc":"Returns keyword list of {Finitomata.Transition.state(), Finitomata.Transition.event()} tuples for determined transition from the current state. The transition is determined, if it is the only transition allowed from the state. Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.Mermaid . parse ( ...&gt; &quot;idle --&gt; |to_s1| s1 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s2| s2 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s3| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |to_s1| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |ambiguous| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |ambiguous| s4 \\n &quot; &lt;&gt; ...&gt; &quot;s3 --&gt; |determined| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s3 --&gt; |determined| s4 \\n &quot; ) ...&gt; Finitomata.Transition . determined ( transitions ) [ s4 : :__end__ , s3 : :determined , idle : :to_s1 ]","ref":"Finitomata.Transition.html#determined/1","title":"Finitomata.Transition.determined/1","type":"function"},{"doc":"Returns {:ok, {event(), state()}} tuple if there is a determined transition from the current state, :error otherwise. The transition is determined, if it is the only transition allowed from the state. Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.Mermaid . parse ( ...&gt; &quot;idle --&gt; |to_s1| s1 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s2| s2 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s3| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |to_s3| s3&quot; ) ...&gt; Finitomata.Transition . determined ( transitions , :s1 ) :error iex&gt; Finitomata.Transition . determined ( transitions , :s2 ) { :ok , { :to_s3 , :s3 } } iex&gt; Finitomata.Transition . determined ( transitions , :s3 ) { :ok , { :__end__ , :* } }","ref":"Finitomata.Transition.html#determined/2","title":"Finitomata.Transition.determined/2","type":"function"},{"doc":"Returns the state after starting one, so-called entry state. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . entry ( transitions ) :s1","ref":"Finitomata.Transition.html#entry/1","title":"Finitomata.Transition.entry/1","type":"function"},{"doc":"Returns true if the state from hsa an outgoing transition with event , false otherwise. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . responds? ( transitions , :s1 , :ok ) true ...&gt; Finitomata.Transition . responds? ( transitions , :s1 , :ko ) false","ref":"Finitomata.Transition.html#responds?/3","title":"Finitomata.Transition.responds?/3","type":"function"},{"doc":"Returns the not ordered list of states, excluding the starting and ending states :* . iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . states ( transitions ) [ :s1 , :s2 ]","ref":"Finitomata.Transition.html#states/1","title":"Finitomata.Transition.states/1","type":"function"},{"doc":"The event in FSM","ref":"Finitomata.Transition.html#t:event/0","title":"Finitomata.Transition.event/0","type":"type"},{"doc":"The state of FSM","ref":"Finitomata.Transition.html#t:state/0","title":"Finitomata.Transition.state/0","type":"type"},{"doc":"The transition is represented by from and to states and the event .","ref":"Finitomata.Transition.html#t:t/0","title":"Finitomata.Transition.t/0","type":"type"},{"doc":"The FSM boilerplate based on callbacks","ref":"readme.html","title":"Finitomata","type":"extras"},{"doc":"Finitomata provides a boilerplate for FSM implementation, allowing to concentrate on the business logic rather than on the process management and transitions/events consistency tweaking. It reads a description of the FSM from a string in PlantUML , Mermaid , or even custom format. Basically, it looks more or less like this PlantUML [ * ] -- &gt; s1 : to_s1 s1 -- &gt; s2 : to_s2 s1 -- &gt; s3 : to_s3 s2 -- &gt; [ * ] : ok s3 -- &gt; [ * ] : ok Mermaid s1 -- &gt; | to_s2 | s2 s1 -- &gt; | to_s3 | s3 Note mermaid does not allow to explicitly specify transitions (and hence event names) from the starting state and to the end state(s), these states names are implicitly set to :* and events to :__start__ and :__end__ respectively. Finitomata validates the FSM is consistent, namely it has a single initial state, one or more final states, and no orphan states. If everything is OK, it generates a GenServer that could be used both alone, and with provided supervision tree. This GenServer requires to implement three callbacks on_transition/4 — mandatory on_failure/3 — optional on_enter/2 — optional on_exit/2 — optional on_terminate/1 — optional on_timer/2 — optional All the callbacks do have a default implementation, that would perfectly handle transitions having a single to state and not requiring any additional business logic attached. Upon start, it moves to the next to initial state and sits there awaiting for the transition request . Then it would call an on_transition/4 callback and move to the next state, or remain in the current one, according to the response. Upon reaching a final state, it would terminate itself. The process keeps all the history of states it went through, and might have a payload in its state.","ref":"readme.html#bird-view","title":"Finitomata - Bird View","type":"extras"},{"doc":"If the event name is ended with a bang (e. g. idle --&gt; |start!| started ) and this event is the only one allowed from this state (there might be several transitions though,) it’d be considered as determined and FSM will be transitioned into the new state instantly. If the event name is ended with a question mark (e. g. idle --&gt; |start?| started ,) the transition is considered as expected to fail; no on_failure/2 callback would be called on failure and no log warning will be printed.","ref":"readme.html#special-events","title":"Finitomata - Special Events","type":"extras"},{"doc":"Recurrent Callback If timer: non_neg_integer() option is passed to use Finitomata , then Finitomata.on_timer/2 callback will be executed recurrently. This might be helpful if FSM needs to update its state from the outside world on regular basis.","ref":"readme.html#fsm-tuning-and-configuration","title":"Finitomata - FSM Tuning and Configuration","type":"extras"},{"doc":"If auto_terminate: true() | state() | [state()] option is passed to use Finitomata , the special __end__ event to transition to the end state will be called automatically under the hood, if the current state is either listed explicitly, or if the value of the parameter is true . Ensuring State Entry If ensure_entry: true() | [state()] option is passed to use Finitomata , the transition attempt will be retried with {:continue, {:transition, {event(), event_payload()}}} message until succeeded. Neither on_failure/2 callback is called nor warning message is logged. The payload would be updated to hold __retries__: pos_integer() key. If the payload was not a map, it will be converted to a map %{payload: payload} .","ref":"readme.html#automatic-fsm-termination","title":"Finitomata - Automatic FSM Termination","type":"extras"},{"doc":"Let’s define the FSM instance defmodule MyFSM do @plantuml &quot;&quot;&quot; [*] --&gt; s1 : to_s1 s1 --&gt; s2 : to_s2 s1 --&gt; s3 : to_s3 s2 --&gt; [*] : ok s3 --&gt; [*] : ok &quot;&quot;&quot; use Finitomata , fsm : @plantuml , syntax : Finitomata.PlantUML ## or uncomment lines below for Mermaid syntax # @mermaid &quot;&quot;&quot; # s1 --&gt; |to_s2| s2 # s1 --&gt; |to_s3| s3 # &quot;&quot;&quot; # use Finitomata, fsm: @mermaid, syntax: Finitomata.Mermaid @impl Finitomata def on_transition ( :s1 , :to_s2 , event_payload , state_payload ) , do : { :ok , :s2 , state_payload } end Now we can play with it a bit. children = [ Finitomata . child_spec ( ) ] Supervisor . start_link ( children , strategy : :one_for_one ) Finitomata . start_fsm MyFSM , &quot;My first FSM&quot; , %{ foo : :bar } Finitomata . transition &quot;My first FSM&quot; , { :to_s2 , nil } Finitomata . state &quot;My first FSM&quot; #⇒ %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}} Finitomata . allowed? &quot;My first FSM&quot; , :* # state #⇒ true Finitomata . responds? &quot;My first FSM&quot; , :to_s2 # event #⇒ false Finitomata . transition &quot;My first FSM&quot; , { :ok , nil } # to final state #⇒ [info] [◉ ⇄] [state: %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}}] Finitomata . alive? &quot;My first FSM&quot; #⇒ false Typically, one would implement all the on_transition/4 handlers, pattern matching on the state/event.","ref":"readme.html#example","title":"Finitomata - Example","type":"extras"},{"doc":"def deps do [ { :finitomata , &quot;~&gt; 0.1&quot; } ] end","ref":"readme.html#installation","title":"Finitomata - Installation","type":"extras"},{"doc":"0.8.0 — :finitomata compiler to warn/hint about not implemented ambiguous transitions 0.7.2 — [FIX] banged! transitions must not be determined 0.6.3 — soft? events which do not call on_failure/2 and do not log errors 0.6.2 — ensure_entry: option to retry a transition 0.6.1 — code cleanup + auto_terminate: option to make :__end__ transition imminent 0.6.0 — on_timer/2 and banged imminent transitions 0.5.2 — state() type on generated FSMs 0.5.1 — fixed specs [credits @egidijusz] 0.5.0 — all callbacks but on_transition/4 are optional, accept impl_for: param to use Finitomata 0.4.0 — allow anonymous FSM instances 0.3.0 — en_entry/2 and on_exit/2 optional callbacks 0.2.0 — Mermaid support Documentation .","ref":"readme.html#changelog","title":"Finitomata - Changelog","type":"extras"},{"doc":"","ref":"fsm.html","title":"Finite Automata","type":"extras"},{"doc":"If you have issues fully understanding Finite Automata aka FSM , here is a one-sentence explanation that might clarify everything. FSM consists of states, transitions, and events, where the state is an adjective (or noun,) the event is a verb and the transition is an adverbial participle. This definition, while being mathematically lax, shallow, cursory, and maybe even perfunctory, reveals the whole and allows to grasp the very core thing about FSM nature: it describes a life in the same way the human language desribes it.","ref":"fsm.html#above-and-beyound","title":"Finite Automata - Above and Beyound","type":"extras"},{"doc":"I’m a big fan of Finite Automata . Always have been. Whenever we deal with a long-lived objects, they eventually have states and FSM does an enormously great job by attesting consistency, eliminating human errors and leaving the implementation with a business logic only. In a mediocre project, fifty if - then - else conditionals might perform as well as one FSM , but unless you are paid for the number of LoCs, FSM is drastically easier to carry on.","ref":"fsm.html#yet-another-library","title":"Finite Automata - Yet Another Library","type":"extras"},{"doc":"This library leverages the power of callbacks to not only completely cover the FSM implementation, but also provide a compile-time proof the FSM is valid and functional. One of the most important things this library provides is the FSM description itself, that is fault-tolerant, not error-prone, and easy to grasp. The FSM definition, which is currently supported in both PlantUML and Mermaid syntaxes, would be drawn in the generated docs of the project using this library. The consumer of this library initiates a transition by calling somewhat like transition(object, event) , then the GenServer does its magic and the callback on_transition/4 gets called. From inside this callback, the consumer implements the business logic and returns the result (the next state to move the FSM to.) There are also syntactic sugar callbacks on_enter/2 , on_exit/2 , on_failure/2 , and on_terminate/1 to allow easy state change reactions, handling of errors, and a final cleanup respectively. All the callbacks do have a default implementation, which would perfectly handle transitions having a single to state and not requiring any additional business logic attached. When needed, this might be turned off. Upon start, FSM moves to its initial state and sits there awaiting for the transition request. Upon this request, it’d call on_transition/4 callback and either move to the next state, or remain in the current one, according to the response from the callback. Upon reaching a final state, it would terminate itself, that’s where on_terminate/1 callback is called from. The process also keeps all the history of states it went through, and might have a payload in its state.","ref":"fsm.html#internals","title":"Finite Automata - Internals","type":"extras"},{"doc":"This library has a compilation-time guarantee the FSM is valid, e. g. has the only one begin state, has at least one end state, all states can be reached, and all the necessary callbacks are defined. That said, if we an FSM has an event initiating transitions from the same state to two different states, and there is no on_transition/4 clause covering that case, the compile-time error would be raised. On the other hande, if the transition is predermined and might lead to the only one state, the callback implementation is not mandatory, because there is no trolley problem between these two states. The FSM definition allows event names, terminated with bangs and/or question marks. If the event name is terminated with a bang ( init! ,) and this event is the only one possible from this state, the event will be called automatically once FSM enters this state. This is handy for moving through initialization or through states which do not require a consumer intervention and might be done immediately after FSM reaches the respective state. If the transition failed in any way (the state has not been left either due to {:error, any()} response received from on_transition/4 or due to other unexpected issue, like if on_transition/4 raised,) the on_failure/2 callback would be called and the warning would be printed to the log. To suppress this behaviour and to allow a transition silently fail, the event should have ended with a question mark ( try_call? .) The event cannot have both a bang and a question mark in its name.","ref":"fsm.html#cool-stuff","title":"Finite Automata - Cool stuff","type":"extras"},{"doc":"Wikipedia provides a turnstile as an example of FSM . Below is the Finitomata take on this FSM . defmodule Turnstile do @fsm &quot;&quot;&quot; built --&gt; |on!| locked locked --&gt; |push| locked locked --&gt; |coin?| unlocked unlocked --&gt; |push| locked unlocked --&gt; |coin?| unlocked unlocked --&gt; |off| destroyed &quot;&quot;&quot; use Finitomata , @fsm , auto_terminate : true def on_transition ( state , :push , _event_payload , state_payload ) do if state == :locked , do : electrocute! ( ) { :ok , :locked , state_payload } end def on_transition ( :locked , :coin? , _event_payload , state_payload ) do { :ok , :unlocked , state_payload } end def on_transition ( :unlocked , :coin , _event_payload , state_payload ) do Logger . info ( &quot;Thanks, this coin will be donated to the animal shelter!&quot; ) { :error , :unexpected_coin } end def on_transition ( _ , :off , _ , state_payload ) , do : { :ok , :destroyed , state_payload } # def on_failure(…), do: … # def on_terminate(…), do: … end The docs for this module would have the following diagram contained (see the ex_doc for how to enable Mermaid .) graph TD built --&gt; |on!| locked locked --&gt; |push| locked locked --&gt; |coin?| unlocked unlocked --&gt; |push| locked unlocked --&gt; |coin?| unlocked unlocked --&gt; |off| destroyed","ref":"fsm.html#wiki-example","title":"Finite Automata - Wiki Example","type":"extras"},{"doc":"Elixir provides an ability to introduce a custom compiler for any mix project. Finitomata leverages this feature, providing a set of diagnostics for modules declaring the FSM .","ref":"compiler.html","title":"Finitomata Compiler","type":"extras"},{"doc":"Consider the following FSM . graph TD idle --&gt; |to_s1!| s1 s1 --&gt; |to_s2| s2 s1 --&gt; |to_s3| s3 s2 --&gt; |to_s1| s3 s2 --&gt; |ambiguous| s3 s2 --&gt; |ambiguous| s4 s3 --&gt; |determined| s3 s3 --&gt; |determined| s4 s4 --&gt; |determined| s4 s4 --&gt; |determined| s5 Here we have transitions which do not require a handler (events to_s1! , to_s1 , to_s2 , and to_s3 define the target state, so just sending the event to the FSM in the respective state would be enough.) Also we have ambiguous and determined events (don’t ask about the name of the latter,) which wouldn’t succeed without the respective Finitomata.on_transition/4 handler declared. This handler must route the event to one of two possible states, conditionally returning {:ok, :s3, state} or {:ok, :s4, state} based on the state, the payload, or the event payload. An attempt to call ambiguous event from state s2 without this handler defined would result in runtime error of the shape { :error , { :ambiguous_transition , { :s2 , :ambiguous } , [ :s3 . :s4 ] } } but we can do better. Here is the custom compiler shining. Let’s see how the editor would have the code highlighted.","ref":"compiler.html#overview","title":"Finitomata Compiler - Overview","type":"extras"},{"doc":"There are three different types of diagnostics provided by :finitomata compiler. Warning Warning is reported both to the language server and to console during project compilation. It says “this module declares some ambiguous transitions not handled in the code.” Here is how it looks in the editor. Info Info is reported in cases when the compiler is not able to determine whether the ambiguous transition has been covered or not. Hint Hint is emitted when the compiler is positive about the ambiguous transition has been covered, just for the sake of better user experience.","ref":"compiler.html#warnings-infos-hints","title":"Finitomata Compiler - Warnings, Infos, Hints","type":"extras"},{"doc":"This compiler is very much WIP , but it’s already fully functional, and, despite it is not able yet to report complicated cases when the transition seems to be handled, but one or more branches are unreachable, it could be a very handy tool for the additional self-check on whether the FSM is covered or the application might suffer the ambiguous , or not reachable , or not allowed runtime errors.","ref":"compiler.html#wip","title":"Finitomata Compiler - WIP","type":"extras"}]